import asyncio
from datetime import datetime
from typing import List, Optional, Sequence, Set

from pydantic import BaseModel

from ..core import BaseClient, with_concurrency_limit
from ..encoders import encode_query_params
from ..types import AccountType, Currency, Impact

PATH = "/users/{user_uuid}/institutions/{institution_id}/transactions"


class TransactionRecord(BaseModel):
    ts: int
    description: Optional[str]
    amount: int
    impact: Impact
    account_number: str
    account_type: Optional[AccountType]
    currency: Optional[Currency]
    labels: Set[str]


class TransactionsMeta(BaseModel):
    page: int
    utc_starttime: Optional[datetime] = None
    utc_endtime: Optional[datetime] = None
    institution_id: str
    user_uuid: str
    total_transactions_count: int
    max_pages: int
    client_uuid: str


class TransactionsAccount(BaseModel):
    account_number: str
    account_type: Optional[AccountType]
    records: List[TransactionRecord]


class TransactionsResponse(BaseModel):
    meta: TransactionsMeta
    transactions: List[TransactionsAccount]

    class Config:
        fields = {"meta": "_meta"}


class BaseTransactionsResource:
    def __init__(self, client: BaseClient):
        self._client = client

    async def _get_page(
        self,
        user_uuid: str,
        institution_id: str,
        utc_starttime: Optional[datetime] = None,
        utc_endtime: Optional[datetime] = None,
        labels: Optional[Sequence[str]] = None,
        page: int = 1,
    ) -> TransactionsResponse:
        async with self._client.session() as session:
            response = await session.get(
                PATH.format(user_uuid=user_uuid, institution_id=institution_id),
                params=encode_query_params(
                    utc_starttime=utc_starttime,
                    utc_endtime=utc_endtime,
                    labels=labels,
                    page=page,
                ),
            )

        assert response.status_code == 200, response.text
        return TransactionsResponse(**response.json())

    async def _get(
        self,
        user_uuid: str,
        institution_id: str,
        utc_starttime: Optional[datetime] = None,
        utc_endtime: Optional[datetime] = None,
        labels: Optional[Sequence[str]] = None,
    ) -> List[TransactionRecord]:
        response = await self._get_page(
            user_uuid=user_uuid,
            institution_id=institution_id,
            utc_starttime=utc_starttime,
            utc_endtime=utc_endtime,
            labels=labels,
        )
        max_pages = response.meta.max_pages
        if max_pages > 1:
            coroutines = [
                self._get_page(
                    user_uuid=user_uuid,
                    institution_id=institution_id,
                    utc_starttime=utc_starttime,
                    utc_endtime=utc_endtime,
                    labels=labels,
                    page=page + 2,
                )
                for page in range(max_pages - 1)
            ]
            response_pages = await asyncio.gather(
                *with_concurrency_limit(coroutines, self._client.concurrency_limit)
            )
            responses = (response, *response_pages)
        else:
            responses = (response,)

        return [
            record
            for response in responses
            for transaction_account in response.transactions
            for record in transaction_account.records
        ]


class AsyncTransactionsResource(BaseTransactionsResource):
    async def get(
        self,
        user_uuid: str,
        institution_id: str,
        utc_starttime: Optional[datetime] = None,
        utc_endtime: Optional[datetime] = None,
        labels: Optional[Sequence[str]] = None,
    ) -> List[TransactionRecord]:
        return await self._get(
            user_uuid=user_uuid,
            institution_id=institution_id,
            utc_starttime=utc_starttime,
            utc_endtime=utc_endtime,
            labels=labels,
        )


class SyncTransactionsResource(BaseTransactionsResource):
    def get(
        self,
        user_uuid: str,
        institution_id: str,
        utc_starttime: Optional[datetime] = None,
        utc_endtime: Optional[datetime] = None,
        labels: Optional[Sequence[str]] = None,
    ) -> List[TransactionRecord]:
        return asyncio.run(
            self._get(
                user_uuid=user_uuid,
                institution_id=institution_id,
                utc_starttime=utc_starttime,
                utc_endtime=utc_endtime,
                labels=labels,
            )
        )



"""
The pngme.api package exposes utilities to interact with Pngme's financial data.
"""


from .client import AsyncClient, Client
from .core import with_concurrency_limit
from .resources.alerts import Alert
from .resources.balances import BalanceRecord
from .resources.credit_report import CreditReport
from .resources.institutions import Institution
from .resources.transactions import TransactionRecord

__all__ = (
    "AsyncClient",
    "Client",
    "with_concurrency_limit",
    "Alert",
    "BalanceRecord",
    "CreditReport",
    "Institution",
    "TransactionRecord",
)



import asyncio
from typing import List

from pydantic import BaseModel

from ..core import BaseClient
from ..types import AccountType, Currency

PATH = "/users/{user_uuid}/institutions"


class InstitutionNames(BaseModel):
    name: str
    display_name: str


class Institution(BaseModel):
    institution_id: str
    institution: InstitutionNames
    currency: Currency
    account_types: List[AccountType]


class InstitutionsMeta(BaseModel):
    user_uuid: str
    client_uuid: str


class InstitutionsResponse(BaseModel):
    meta: InstitutionsMeta
    institutions: List[Institution]

    class Config:
        fields = {"meta": "_meta"}


class BaseInstitutionsResource:
    def __init__(self, client: BaseClient):
        self._client = client

    async def _get(self, user_uuid: str) -> List[Institution]:
        async with self._client.session() as session:
            response = await session.get(PATH.format(user_uuid=user_uuid))

        assert response.status_code == 200, response.text
        return InstitutionsResponse(**response.json()).institutions


class AsyncInstitutionsResource(BaseInstitutionsResource):
    async def get(self, user_uuid: str) -> List[Institution]:
        return await self._get(user_uuid)


class SyncInstitutionsResource(BaseInstitutionsResource):
    def get(self, user_uuid: str) -> List[Institution]:
        return asyncio.run(self._get(user_uuid))



from datetime import datetime
from typing import Any, Dict


def _to_isoformat(value: datetime) -> str:
    """Cast datetime to string removing microseconds."""
    return value.replace(microsecond=0).isoformat()


def _filter_truthy_values(**kwargs: Any) -> Dict[str, Any]:
    """Return a copy containing only keys containing truthy values."""
    return {k: v for k, v in kwargs.items() if v}


def encode_query_params(**kwargs: Any) -> Dict[str, Any]:
    """Serialize incompatible types and remove falsy values."""
    query_params = {}
    for key, value in _filter_truthy_values(**kwargs).items():
        if isinstance(value, datetime):
            query_params[key] = _to_isoformat(value)
        else:
            query_params[key] = value
    return query_params



from .core import BaseClient
from .resources.alerts import AsyncAlertsResource, SyncAlertsResource
from .resources.balances import AsyncBalancesResource, SyncBalancesResource
from .resources.credit_report import AsyncCreditReportResource, SyncCreditReportResource
from .resources.institutions import AsyncInstitutionsResource, SyncInstitutionsResource
from .resources.transactions import AsyncTransactionsResource, SyncTransactionsResource
from .resources.users import AsyncUsersResource, SyncUsersResource


class AsyncClient(BaseClient):
    @property
    def alerts(self) -> AsyncAlertsResource:
        return AsyncAlertsResource(self)

    @property
    def balances(self) -> AsyncBalancesResource:
        return AsyncBalancesResource(self)

    @property
    def credit_report(self) -> AsyncCreditReportResource:
        return AsyncCreditReportResource(self)

    @property
    def institutions(self) -> AsyncInstitutionsResource:
        return AsyncInstitutionsResource(self)

    @property
    def transactions(self) -> AsyncTransactionsResource:
        return AsyncTransactionsResource(self)

    @property
    def users(self) -> AsyncUsersResource:
        return AsyncUsersResource(self)


class Client(BaseClient):
    @property
    def alerts(self) -> SyncAlertsResource:
        return SyncAlertsResource(self)

    @property
    def balances(self) -> SyncBalancesResource:
        return SyncBalancesResource(self)

    @property
    def credit_report(self) -> SyncCreditReportResource:
        return SyncCreditReportResource(self)

    @property
    def institutions(self) -> SyncInstitutionsResource:
        return SyncInstitutionsResource(self)

    @property
    def transactions(self) -> SyncTransactionsResource:
        return SyncTransactionsResource(self)

    @property
    def users(self) -> SyncUsersResource:
        return SyncUsersResource(self)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['api', 'api.resources']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.22.0,<0.23.0',
 'nest-asyncio>=1.5.4,<2.0.0',
 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'pngme-api',
    'version': '0.0.9',
    'description': "API client used to access Pngme's financial data APIs.",
    'long_description': '<p align="center">\n  <img src="https://admin.pngme.com/logo.png" alt="Pngme" width="100" height="100">\n</p>\n\n<h3 align="center">Python API Client</h3>\n\nThis package exposes a synchronous and asynchronous client used to interact with Pngme\'s financial data APIs.\n\n## Install\n\nInstall the latest version with:\n\n```bash\npip3 install pngme-api\n```\n\n## Quick start\n\nCreate a `Client` instance using your API `token` found in the [Pngme Dashboard](https://admin.pngme.com):\n\n```python\nfrom pngme.api import Client\n\ntoken = "" # your API token\nclient = Client(token)\n```\n\n> If you\'re using [`asyncio`](https://docs.python.org/3/library/asyncio.html), you can import and use the `AsyncClient` instead.\n\nWe can list or search the available [`/users`](https://developers.api.pngme.com/reference/get_users):\n\n```python\nusers = client.users.get()\nusers = client.users.get(search="2343456789012")\n```\n\nFor a user of interest, we can get a list of the user\'s [`/institutions`](https://developers.api.pngme.com/reference/get_users-user-uuid-institutions):\n\n```python\nuser_uuid = "33b6215d-3d75-4271-801c-6da27603a8be"\n\ninstitutions = client.institutions.get(user_uuid=user_uuid)\n```\n\nThen for a given institution, we can get a list of the user\'s [`/transactions`](https://developers.api.pngme.com/reference/get_users-user-uuid-institutions-institution-id-transactions), [`/balances`](https://developers.api.pngme.com/reference/get_users-user-uuid-institutions-institution-id-balances), or [`/alerts`](https://developers.api.pngme.com/reference/get_users-user-uuid-institutions-institution-id-alerts):\n\n```python\nuser_uuid = "33b6215d-3d75-4271-801c-6da27603a8be"\ninstitution_id = "zenithbank"\n\ntransactions = client.transactions.get(user_uuid=user_uuid, institution_id=institution_id)\nbalances = client.balances.get(user_uuid=user_uuid, institution_id=institution_id)\nalerts = client.alerts.get(user_uuid=user_uuid, institution_id=institution_id)\n```\n\n## Development environment\n\nWe use [Poetry](https://python-poetry.org) to build, package, and publish this project. You\'ll need to [install Poetry](https://python-poetry.org/docs/#installation) to use the development tools.\n\nClone this repo and install the development dependencies:\n\n```bash\nmake install\n```\n\n> You may need to configure your IDE to point to the newly created virtual environment after running install.\n\nThis will create a virtual environment in the `.venv` directory, which is the convention Pipenv and poetry both use for in-project virtual environments.\n\nYou can type `make help` to see a list of other options, which aren\'t strictly necessary as part of our day-to-day development practices.\n\n### Integration tests\n\nYou can replicate the integration tests (which run in a GitHub action on every push) locally:\n\n```bash\nmake ci\n```\n\n### Dependencies\n\nProduction dependencies are listed in [setup.cfg](./setup.cfg) under `options.install_requires`, see [`setuptools` dependency management](https://setuptools.pypa.io/en/latest/userguide/quickstart.html#dependency-management).\n\nDevelopment dependencies are listed in [requirements.txt](./requirements.txt).\n\nRunning `make install` will install **both** production and development dependencies. Running `pip install .` (where `.` is the path to the directory containing [pyproject.toml](./pyproject.toml)) will install only the production dependencies.\n',
    'author': 'Ben Fasoli',
    'author_email': 'ben@pngme.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://developers.api.pngme.com/reference/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
